#line 27 "heap.nw"
/* booleans */
#ifndef __cplusplus
typedef enum bool { false, true } boolean;
#else
# define boolean bool
#endif /*__cplusplus*/

/* words */
#define word_size	sizeof(word)
#define wordsof(x)	((sizeof(x) + word_size - 1) / word_size)

typedef void *word;

#line 55 "heap.nw"
/* check for boxed-ness of a node */
#define is_unboxed(node)        (!ONLY_BOXED_OBJECTS && ((long)(node) & 0x01))
#define is_boxed(node)          !is_unboxed(node)

#if !ONLY_BOXED_OBJECTS
# define mk_unboxed(x)		(Node *)(((x) << 1) | 0x01)
# define unboxed_val(node)	((long)(node) >> 1)
#endif

#line 83 "heap.nw"
enum {
    CAPP_KIND,			/* constructor application */
    PAPP_KIND,			/* partial application node */
    LVAR_KIND,			/* (unbound) logic variable */
    LAZY_KIND,			/* lazy function application node */
    INDIR_KIND,			/* indirection node */
    VECTOR_KIND,		/* vector nodes */
    CHAR_KIND,			/* character node */
    INT_KIND,			/* integer number */
    FLOAT_KIND,			/* floating-point number */
    PTR_KIND,			/* foreign pointer */
    FOREIGN_PTR_KIND,		/* finalized foreign pointer */
    ABSTRACT_KIND,		/* external objects */
    THREAD_KIND,		/* thread queue node */
    SEARCH_CONT_KIND,		/* search continuation */
    SCRIPT_KIND			/* saved script/trail */
};

#line 154 "heap.nw"
/* node types -- defined below */
typedef union node	       Node;
typedef struct node_info       NodeInfo;
typedef union thread_node      *ThreadQueue;	/* defined in threads.h */
typedef struct search_space    SearchSpace;	/* defined in spaces.h */
typedef struct constraint      Constraint;
typedef void (*FinalFun)(Node *);

struct node_info
{
    const short	       kind;		/* node kind */
    const short	       tag;		/* tag value */
    const unsigned int length;		/* length of the node */
    const int          *offset_table;	/* pointer offset table */
    const char	       *cname;		/* closure name */
    const Label	       eval;		/* entry point for evaluation */
    const Label	       entry;		/* function entry point */
    const FinalFun     final_fun;	/* optional finalization function */
};

#define node_kind(node)		(node)->info->kind
#define node_tag(node)		(node)->info->tag
#define node_size(node)		(node)->info->length
#define is_capp_node(node)	(node_kind(node) == CAPP_KIND)
#define is_papp_node(node)	(node_kind(node) == PAPP_KIND)
#define is_variable_node(node)	(node_kind(node) == LVAR_KIND)
#define is_lazy_node(node)	(node_kind(node) == LAZY_KIND)
#define is_indir_node(node)	(node_kind(node) == INDIR_KIND)
#define is_vector_node(node)	(node_kind(node) == VECTOR_KIND)
#define is_char_node(node)	(node_kind(node) == CHAR_KIND)
#define is_int_node(node)	(node_kind(node) == INT_KIND)
#define is_float_node(node)	(node_kind(node) == FLOAT_KIND)
#define is_ptr_node(node)	(node_kind(node) == PTR_KIND)
#define is_foreign_ptr_node(node) (node_kind(node) == FOREIGN_PTR_KIND)
#define is_abstract_node(node)	(node_kind(node) == ABSTRACT_KIND)
#define is_search_cont_node(node) (node_kind(node) == SEARCH_CONT_KIND)

#define word_offsetof(type,field) ((word *)&(((type *)0)->field) - (word *)0)

extern const int ot_binary[];

#line 205 "heap.nw"
struct closure_node {
    NodeInfo *info;
    Node     *args[1];
};

#define closure_argc(node)	(node_size(node) - closure_node_size(0))
#define closure_node_size(argc)	(wordsof(struct closure_node) + (argc) - 1)

#line 220 "heap.nw"
struct vector_node {
    NodeInfo *info;
    unsigned length;
    Node     *args[1];
};

#define vector_argc(node) \
    (((struct vector_node *)(node))->length - vector_node_size(0))
#define vector_node_size(argc) \
    (wordsof(struct vector_node) + (argc) - 1)

#line 240 "heap.nw"
#define is_large_int(i) \
    (ONLY_BOXED_OBJECTS || (i) < LONG_MIN/2 || (i) > LONG_MAX/2)

extern NodeInfo int_info;
struct int_node {
    NodeInfo *info;
    long     i;
};
#define int_node_size	wordsof(struct int_node)

#if ONLY_BOXED_OBJECTS
# define long_val(node)	(node)->i.i
#else
# define long_val(node)	(is_unboxed(node) ? unboxed_val(node) : (node)->i.i)
#endif

#line 279 "heap.nw"
extern NodeInfo float_info;
struct float_node {
    NodeInfo *info;
    double   d;
};
#define float_node_size         wordsof(struct float_node)

#if UNALIGNED_DOUBLE
# define get_double_val(_d,node)	_d = (node).d
# define put_double_val(node,_d)	(node).d = _d
#else
union _conv {
    double d;
    long   l[2];
};

# define get_double_val(_d,node) do { \
    union _conv conv; \
    conv.l[0] = ((long *)&(node).d)[0]; \
    conv.l[1] = ((long *)&(node).d)[1]; \
    _d = conv.d; \
} while ( 0 )
# define put_double_val(node,_d) do { \
    union _conv conv; \
    conv.d = _d; \
    ((long *)&(node).d)[0] = conv.l[0]; \
    ((long *)&(node).d)[1] = conv.l[1]; \
} while ( 0 )
#endif

#line 321 "heap.nw"
extern NodeInfo char_info;
struct char_node {
    NodeInfo *info;
    int	     ch;
};

#define char_node_size		wordsof(struct char_node)

extern struct char_node char_table[256];
extern void init_chars(void);

#line 363 "heap.nw"
enum { PTR_TAG, FUNPTR_TAG, STABPTR_TAG };

extern NodeInfo ptr_info, funptr_info, stabptr_info;
struct ptr_node {
    NodeInfo *info;
    void     *ptr;
};
#define ptr_node_size		wordsof(struct ptr_node)

extern const struct ptr_node null_ptr, null_funptr, null_stabptr;

#line 402 "heap.nw"
struct foreign_ptr_node {
    NodeInfo		 *info;
    SearchSpace		 *spc;
    void		 *ptr;
    union finalizer_list *fptr;
};
#define foreign_ptr_node_size	wordsof(struct foreign_ptr_node)

#line 424 "heap.nw"
extern NodeInfo indir_info;
struct indir_node {
    NodeInfo	*info;
#if !COPY_SEARCH_SPACE
    SearchSpace	*spc;		/* defining search space (of original node) */
#endif
    Node	*node;		/* referenced node */
};
#define indir_node_size		wordsof(struct indir_node)

#line 448 "heap.nw"
extern NodeInfo variable_info_table[];
struct variable_node {
    NodeInfo	*info;
    SearchSpace	*spc;		/* defining search space */
    ThreadQueue	wq;		/* wait queue */
    Constraint	*cstrs;		/* constraint list */
};
#define variable_node_size      wordsof(struct variable_node)

#line 481 "heap.nw"
extern NodeInfo constraint_info;
struct constraint {
    NodeInfo   *info;
    Constraint *cstrs;		/* link to next constraint or 0 */
};

#line 510 "heap.nw"
enum { NOUPD_TAG, UPD_TAG, QUEUEME_TAG };

#define is_unshared_lazy_node(node) \
    (is_lazy_node(node) && node_tag(node) == NOUPD_TAG)
#define is_shared_lazy_node(node) \
    (is_lazy_node(node) && node_tag(node) == UPD_TAG)
#define is_queueMe_node(node) \
    (is_lazy_node(node) && node_tag(node) == QUEUEME_TAG)

#line 532 "heap.nw"
struct suspend_node {
    NodeInfo	*info;
    SearchSpace	*spc;		/* defining search space */
    Node	*args[1];	/* arguments */
};
#define	suspend_argc(node)	(node_size(node) - raw_suspend_node_size(0))
#define suspend_node_size(argc)	raw_suspend_node_size((argc) > 1 ? (argc) : 1)
#define raw_suspend_node_size(argc) (wordsof(struct suspend_node) + (argc) - 1)

#line 549 "heap.nw"
extern NodeInfo queueMe_info_table[];
struct queueMe_node {
    NodeInfo	*info;
    SearchSpace	*spc;		/* defining search space */
    ThreadQueue wq;		/* pointer to waitqueue */
};
#define queueMe_node_size       wordsof(struct queueMe_node)

#line 574 "heap.nw"
extern NodeInfo search_cont_info;
struct search_cont_node {
    NodeInfo	*info;
    SearchSpace *spc;		/* local space of the continuation */
    Node	*susp;		/* suspended goal application */
    Node	*var;		/* goal variable */
    Label	code;		/* next instruction to be executed */
    ThreadQueue rq;		/* saved ready queue */
};
#define search_cont_node_size   wordsof(struct search_cont_node)

#line 595 "heap.nw"
union node {
    NodeInfo *info;
    struct closure_node c;
    struct vector_node a;
    struct char_node ch;
    struct int_node i;
    struct float_node f;
    struct ptr_node p;
    struct foreign_ptr_node fp;
    struct variable_node v;
    struct suspend_node s;
    struct queueMe_node q;
    struct indir_node n;
    struct search_cont_node sc;
};

#line 618 "heap.nw"
#define DECLARE_CONST(name) \
DECLARE_DATA(name); \
extern NodeInfo *const name##_node

#define DECLARE_DATA(name) \
extern NodeInfo name##_info

#define DEFINE_CONST(name,cname,tag) \
DEFINE_DATA(name,cname,tag,0); \
NodeInfo *const name##_node = &name##_info

#define DEFINE_DATA(name,cname,tag,arity) \
NodeInfo name##_info = { \
    CAPP_KIND, tag, closure_node_size(arity), 0, cname, eval_whnf, 0, 0 \
}

#line 640 "heap.nw"
#define CONST(name)		(Node *)(&name##_node)
#define INFO(name)		&name##_info

#line 654 "heap.nw"
#define FUNINFO(fname,entrypoint,arity) \
{ LAZY_KIND, NOUPD_TAG, closure_node_size(arity), 0, fname, \
  eval_clos_##arity, entrypoint, 0 }

#define PAPPINFO(fname,argc,entrypoint,arity) \
{ PAPP_KIND, (arity) - (argc), closure_node_size(argc), 0, fname, \
  eval_whnf, entrypoint, 0 },

#line 669 "heap.nw"
#if !COPY_SEARCH_SPACE
# define LAZYINFO(fname,entrypoint,arity) { \
    { LAZY_KIND, UPD_TAG, suspend_node_size(arity), 0, fname, \
      eval_lazy_##arity, entrypoint, 0 }, \
    { LAZY_KIND, QUEUEME_TAG, suspend_node_size(arity), 0, 0, \
      eval_queueMe, 0, 0 },					\
    { INDIR_KIND, 0, suspend_node_size(arity), 0, 0, eval_indir, 0, 0 } \
}
#else
# define LAZYINFO(fname,entrypoint,arity) { \
    { LAZY_KIND, UPD_TAG, suspend_node_size(arity), 0, fname, \
      eval_lazy_##arity, entrypoint, 0 } \
}
#endif

#line 698 "heap.nw"
extern boolean is_tuple(NodeInfo *info);
extern boolean is_operator(NodeInfo *info);

#line 751 "heap.nw"
extern word *hp, *hlim;
extern word *heap_base, *heap_end;

#line 773 "heap.nw"
extern boolean in_gc;
extern void collect(unsigned int);
#define CHECK_HEAP(n) do { if ( hp + (n) > heap_end ) collect(n); } while ( 0 )

#line 787 "heap.nw"
#define primPerformGC()	collect(0)

#line 805 "heap.nw"
struct root {
    struct root *next;
    Node	**root;
};

extern struct root *global_roots, *local_roots;

#line 826 "heap.nw"
#define BEGIN_LOCAL_ROOTS1(x) { \
    LOCAL_ROOTS1(x);		\
    local_roots = &__local_root__;
#define BEGIN_LOCAL_ROOTS2(x,y) { \
    LOCAL_ROOTS2(x,y);		  \
    local_roots = &__local_root2_;
#define BEGIN_LOCAL_ROOTS3(x,y,z) { \
    LOCAL_ROOTS3(x,y,z);	    \
    local_roots = &__local_root3_;
#define END_LOCAL_ROOTS		       \
    local_roots = __local_root__.next; \
}

#define LOCAL_ROOTS1(x)		       \
    struct root __local_root__;	       \
    __local_root__.next = local_roots; \
    __local_root__.root = (Node **)&(x)
#define LOCAL_ROOTS2(x,y)		   \
    struct root __local_root2_;		   \
    LOCAL_ROOTS1(x);			   \
    __local_root2_.next = &__local_root__; \
    __local_root2_.root = (Node **)&(y)
#define LOCAL_ROOTS3(x,y,z)		   \
    struct root __local_root3_;		   \
    LOCAL_ROOTS2(x,y);			   \
    __local_root3_.next = &__local_root2_; \
    __local_root3_.root = (Node **)&(z)

#line 862 "heap.nw"
extern void add_global_root(Node **root);
extern void remove_global_root(Node **root);

#line 906 "heap.nw"
extern void register_final(Node *node);

#line 916 "heap.nw"
extern void release_mem(void);

#line 925 "heap.nw"
extern void init_heap(unsigned long);

#line 932 "heap.nw"
extern void heap_exhausted(void) __attribute__ ((noreturn));

